import os
from io import BytesIO
from pathlib import Path

import numpy as np
import rasterio
from PIL import Image
from rasterio.control import GroundControlPoint

TILE_SIZE = 256


def create_tiff(
    tiles_directory: Path,
    layers_bounds: list[tuple[int, int, int, int]],
    zoom: int,
    width: int,
    height: int,
    gcps: list[GroundControlPoint],
) -> BytesIO:
    layer_bounds = layers_bounds[zoom]
    width_in_tiles = layer_bounds[2] - layer_bounds[0] + 1
    height_in_tiles = layer_bounds[3] - layer_bounds[1] + 1

    total_width = width_in_tiles * TILE_SIZE
    total_height = height_in_tiles * TILE_SIZE

    # RGBA image
    data = np.zeros((total_height, total_width, 4), dtype=np.uint8)
    data[:, :, 0:3] = 255  # white background
    data[:, :, 3] = 0  # alpha=0 transparent

    for x in range(width_in_tiles):
        for y in range(height_in_tiles):
            tile_filename = f"{tiles_directory}/{zoom}_{x}_{y}.webp"

            if os.path.exists(tile_filename):
                tile_img = Image.open(tile_filename).convert("RGBA")
                tile_array = np.array(tile_img)
                y_offset = y * TILE_SIZE
                x_offset = x * TILE_SIZE
                data[
                    y_offset : y_offset + TILE_SIZE, x_offset : x_offset + TILE_SIZE
                ] = tile_array

    # Crop image to the actual size
    data = data[:height, :width]

    # Write new GeoTIFF to a BytesIO object
    buffer = BytesIO()

    with rasterio.open(
        buffer,
        "w",
        driver="GTiff",
        height=height,
        width=width,
        count=4,
        dtype="uint8",
        gcps=gcps,
        crs="EPSG:4326",
        compress="LZW",
    ) as dst:
        dst.write(data[:, :, 0], 1)  # Red channel
        dst.write(data[:, :, 1], 2)  # Green channel
        dst.write(data[:, :, 2], 3)  # Blue channel
        dst.write(data[:, :, 3], 4)  # Alpha channel

    buffer.seek(0)

    return buffer



from rasterio.control import GroundControlPoint


def get_gcps(
    corner_coords: list[tuple[float, ...]],
    image_width: int,
    image_height: int,
) -> list[GroundControlPoint]:
    """
    Create Ground Control Points (GCPs) for georeferencing an image.

    Args:
        corner_coords (list[tuple[float, float]]): Geographic coordinates of the image corners.
        image_width (int): Image width in pixels.
        image_height (int): Image height in pixels.

    Returns:
        list[GroundControlPoint]: A list of GCPs mapping image corners to geographic
    """

    # Define image coordinates for the corners (row, col) starting from top-left
    image_coords = [
        (0, 0),  # Upper-left (UL)
        (0, image_width - 1),  # Upper-right (UR)
        (image_height - 1, image_width - 1),  # Lower-right (LR)
        (image_height - 1, 0),  # Lower-left (LL)
    ]

    # Create GCPs (Ground Control Points) mapping image corners to geographic coordinates
    return [
        GroundControlPoint(
            row=image_coords[i][0],
            col=image_coords[i][1],
            x=corner_coords[i][0],  # Longitude
            y=corner_coords[i][1],  # Latitude
            z=0,  # Elevation (optional, usually 0 for 2D)
        )
        for i in range(4)
    ]



import struct
import sys
from pathlib import Path
from pprint import pprint
from typing import BinaryIO

TAG_LENGTH = 4
WEBP_TAG = b"RIFF"

INT_DATA = 0xFFFFFFFE
KNOWN_TAGS = [
    b"PGD_CO",
    b"PGD_CB",
    b"PGD_CT",
    b"SHFTPROP",
    b"SHFTPPRO",
    b"source",
    b"creator",
    b"PPTY",
    b"PPTX",
    b"image_width_px",
    b"image_height_px",
    b"PGD_RASTER_OSF",
    b"import_time",
    b"lat-0",
    b"lon-0",
    b"lat-1",
    b"lon-1",
    b"lat-2",
    b"lon-2",
    b"lat-3",
    b"lon-3",
    b"PGD_RASTER_USF",
]


def get_tag_at_offset(
    data: bytes, offset: int
) -> tuple[bytes | None, int | str | None, int]:
    for tag in KNOWN_TAGS:
        # Convert the tag to bytes
        tag_len = len(tag)

        # Check if the tag is found at the current offset
        if data[offset : offset + tag_len] == tag:
            # Move offset past the tag
            offset += tag_len

            # Read the number of bytes following the tag
            if offset + TAG_LENGTH > len(data):
                raise ValueError("Unexpected end of file while reading length.")

            data_length = struct.unpack(">I", data[offset : offset + TAG_LENGTH])[0]
            offset += TAG_LENGTH

            if data_length == INT_DATA:
                data_length = 8
                tag_data = data[offset : offset + data_length]
                tag_data = int.from_bytes(tag_data, "big")

            else:
                tag_data = data[offset : offset + data_length]
                tag_data = tag_data.decode("utf-8")

            offset += data_length

            # Read the next `length` bytes for the tag's data
            if offset > len(data):
                print("Unexpected end of file while reading data.")
                print(f"Tag: {tag}")
                break

            return tag, tag_data, offset

    return None, None, offset + 1


def get_metadata(file: BinaryIO) -> list[dict[bytes, int | str]]:
    result = [dict()]
    level = 0

    # TODO avoid reading the whole file into memory
    file.seek(0)
    data = file.read()

    max_offset = data.find(WEBP_TAG) or len(data)
    offset = 0

    while offset < max_offset:
        try:
            tag, tag_data, offset = get_tag_at_offset(data, offset)

        except Exception as e:
            print(f"Error at offset {offset}: {e}")
            tag, tag_data, offset = None, None, offset + 1

        if tag is not None:
            # Convert the data to a readable string and append to results
            if tag == b"PGD_CO" and tag in result[level]:
                level += 1
                result.append(dict())

            result[level][tag] = tag_data

    return result


if __name__ == "__main__":
    file_path = Path(sys.argv[1])

    with open(file_path, "rb") as file:
        parsed_data = get_metadata(file)

        for level, data in enumerate(parsed_data):
            print(f"Level {level}:")
            pprint(data)
            print()



import sys
from pathlib import Path

import matplotlib.font_manager as fm
import numpy as np
import rasterio
from affine import Affine
from geopandas import GeoDataFrame
from PIL import Image, ImageDraw, ImageFont
from pykml import parser
from rasterio.features import rasterize
from rasterio.transform import from_bounds
from shapely.geometry import Polygon, Point

FONT_PATH = fm.findfont(fm.FontProperties(family="sans serif"))
MAX_FONT_SIZE = 100


def color_to_rgb(color: str) -> tuple[int, int, int]:
    """
    Convert a color string to RGB values.
        "color": A color string in the format "#RRGGBB"
    Returns:
        A tuple of red, green, and blue values (0-255)
    """

    return int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16)


def _kml_placemarks_to_geodataframe(placemarks: list) -> GeoDataFrame:
    data = []

    for placemark in placemarks:
        try:
            name = placemark.name.text

            # Default color
            color = "000000"

            # Get color from LineStyle
            if hasattr(placemark, "Style") and hasattr(placemark.Style, "LineStyle"):
                color = placemark.Style.LineStyle.color.text

            children = placemark.getchildren()
            if hasattr(placemark, "MultiGeometry"):
                children = placemark.MultiGeometry.getchildren()

            for geometry in children:
                if "Point" in geometry.tag:
                    coords = geometry.coordinates.text.split(",")
                    shape = Point(float(coords[0]), float(coords[1]))

                elif "LineString" in geometry.tag:
                    coords_text = geometry.coordinates.text.split()
                    coords = [
                        list(map(float, coord.split(",")[:2])) for coord in coords_text
                    ]
                    coords = coords + coords[::-1]  # Close the LineString
                    shape = Polygon(coords)

                elif "Polygon" in geometry.tag:
                    coords = (
                        geometry.outerBoundaryIs.LinearRing.coordinates.text.split()
                    )
                    shape = Polygon(
                        [map(float, coord.split(",")[:2]) for coord in coords]
                    )

                else:
                    continue

                data.append(
                    {
                        "name": name,
                        "geometry": shape,
                        "color": f"#{color}",
                    }
                )

        except Exception as e:
            print(f"Unable to parse placemark: {placemark.__dict__}")
            print(e)

    gdf = GeoDataFrame(data)

    # Add WGS 84 CRS
    gdf.crs = "EPSG:4326"

    # Transform geometries to EPSG:3857
    gdf.to_crs("EPSG:3857", inplace=True)

    return gdf


def _parse_kml_to_geodataframe(kml_file: Path) -> GeoDataFrame:
    with open(kml_file) as file:
        doc = parser.parse(file)

    root = doc.getroot()
    placemarks = root.findall(".//{http://www.opengis.net/kml/2.2}Placemark")

    return _kml_placemarks_to_geodataframe(placemarks)


def _draw_text_labels(
    image: Image.Image,
    gdf: GeoDataFrame,
    transform: Affine,
    colum_name: str = "name",
    font_size: int = 24,
    color: str | None = None,
    offset: tuple[int, int] = (0, 0),
):
    """
    Draw text labels from a GeoDataFrame onto a raster image.

    Args:
        image: A Pillow Image object
        gdf: A GeoDataFrame containing geometries to draw and the text column
        colum_name: The column name containing text to draw
        font_size: The font size for the text labels
        color: The color for the text labels in the format "#RRGGBB"
        offset: A tuple of x and y pixel offsets for the text labels

    Returns:
        None
    """

    draw = ImageDraw.Draw(image)

    for _, row in gdf.iterrows():
        # Calculate text position (centroid of geometry)
        centroid = row.geometry.centroid
        text = str(row[colum_name])
        x, y = ~transform * (centroid.x, centroid.y)
        x += offset[0]
        y += offset[1]

        if color is not None:
            red, green, blue = color_to_rgb(color)
        else:
            red, green, blue = color_to_rgb(str(row["color"]))

        # Adjust font size based on geometry width
        font = ImageFont.truetype(FONT_PATH, MAX_FONT_SIZE)
        font = ImageFont.truetype(FONT_PATH, font_size)
        text_width = font.getlength(text)

        draw.text(
            (x - text_width // 2, y - font.size // 2),
            text,
            fill=(red, green, blue, 255),
            font=font,
            stroke_width=0.1,
            stroke_fill=(0, 0, 0, 255),
        )


def _draw_geomteries(
    gdf: GeoDataFrame,
    resolution: float = 0.5,
    outline_width: int = 4,
) -> Image.Image:
    """
    Draw geometries from a GeoDataFrame onto a raster image.

    Args:
        gdf: A GeoDataFrame containing geometries to draw and the color column
             with hex colors in the format "#RRGGBB"
        resolution: The output raster resolution in meters
        outline_width: The width of the geometry outline in pixels

    Returns:
        An RGBA raster image
    """

    # Define the output raster dimensions and bounds
    buffer_width = outline_width * resolution / 2
    boundary = gdf.geometry.buffer(buffer_width)
    bounds = boundary.total_bounds
    width = int((bounds[2] - bounds[0]) / resolution)
    height = int((bounds[3] - bounds[1]) / resolution)

    # Define the affine transformation for the raster
    transform: Affine = from_bounds(
        bounds[0], bounds[1], bounds[2], bounds[3], width, height
    )
    assert isinstance(transform, Affine)

    # Create empty arrays for RGB and Alpha channels
    r_channel = np.zeros((height, width), dtype=np.uint8)
    g_channel = np.zeros((height, width), dtype=np.uint8)
    b_channel = np.zeros((height, width), dtype=np.uint8)
    alpha_channel = np.zeros((height, width), dtype=np.uint8)

    # Group geometries by color
    grouped = gdf.groupby("color")

    # Rasterize each color group
    for color, group in grouped:
        red, green, blue = color_to_rgb(str(color))
        shapes = (
            (geom.buffer(buffer_width), 1)
            if geom.geom_type == "Point"
            else (geom.boundary.buffer(buffer_width), 1)
            for geom in group.geometry
        )
        mask = rasterize(
            shapes,
            out_shape=(height, width),
            transform=transform,
            fill=0,
            dtype=np.uint8,
        )

        # Apply color and set alpha for geometries
        r_channel[mask == 1] = red
        g_channel[mask == 1] = green
        b_channel[mask == 1] = blue
        alpha_channel[mask == 1] = 255  # Fully opaque for geometries

    # Create an RGBA image using Pillow for adding text
    rgba_image = np.stack([r_channel, g_channel, b_channel, alpha_channel], axis=-1)

    image = Image.fromarray(rgba_image, mode="RGBA")
    image.info["transform"] = transform
    return image


# Step 2: Rasterize geometries into a GeoTIFF
def _geodataframe_to_geotiff(
    gdf: GeoDataFrame,
    output_file: Path,
    resolution: float = 0.5,
    outline_width: int = 4,
    font_size: int = 24,
    font_color: str | None = None,
    offset: tuple[int, int] = (0, 0),
):
    image = _draw_geomteries(gdf, resolution, outline_width)

    _draw_text_labels(
        image,
        gdf,
        image.info["transform"],
        font_size=font_size,
        color=font_color,
        offset=offset,
    )

    # Write RGB raster to a GeoTIFF
    with rasterio.open(
        output_file,
        "w",
        driver="GTiff",
        height=image.height,
        width=image.width,
        count=4,
        dtype="uint8",
        crs=gdf.crs,
        transform=image.info["transform"],
        compress="LZW",
    ) as dst:
        dst.write(np.array(image)[:, :, 0], 1)  # Red channel
        dst.write(np.array(image)[:, :, 1], 2)  # Green channel
        dst.write(np.array(image)[:, :, 2], 3)  # Blue channel
        dst.write(np.array(image)[:, :, 3], 4)  # Alpha channel

        print(f"GeoTIFF saved as {output_file}")


def kml_to_geotiff(
    kml_file: Path,
    output_file: Path,
    resolution: float = 0.5,
    outline_width: int = 4,
    font_size: int = 24,
    font_color: str | None = None,
    offset: tuple[int, int] = (0, 0),
):
    """
    Convert a KML file to a GeoTIFF file with geometries and labels.
    """

    gdf = _parse_kml_to_geodataframe(kml_file)

    _geodataframe_to_geotiff(
        gdf, output_file, resolution, outline_width, font_size, font_color, offset
    )


def placemarks_to_geotiff(
    placemarks: list,
    output_file: Path,
    resolution: float = 0.5,
    outline_width: int = 4,
    font_size: int = 24,
    font_color: str | None = None,
    offset: tuple[int, int] = (0, 0),
):
    """
    Convert a list of KML placemarks to a GeoTIFF file with geometries and labels.
    """

    gdf = _kml_placemarks_to_geodataframe(placemarks)

    _geodataframe_to_geotiff(
        gdf, output_file, resolution, outline_width, font_size, font_color, offset
    )


if __name__ == "__main__":
    kml_file = Path(sys.argv[1])

    if len(sys.argv) > 2:
        output_file = Path(sys.argv[2])
    else:
        output_file = kml_file.with_suffix(".tif")

    kml_to_geotiff(kml_file, output_file)



from typing import Generator


def get_tiles(
    layers_bounds: list[tuple[int, int, int, int]],
) -> Generator[tuple[int, int, int], None, None]:
    def process_tile(
        layer: int, x: int, y: int
    ) -> Generator[tuple[int, int, int], None, None]:
        if layer < 0 or layer >= len(layers_bounds):
            return

        layer_bounds = layers_bounds[layer]

        if x >= layer_bounds[2] or y >= layer_bounds[3]:
            return

        if x < layer_bounds[0] or y < layer_bounds[1]:
            return

        for dy in range(2):
            for dx in range(2):
                yield from process_tile(layer + 1, x * 2 + dx, y * 2 + dy)

        yield (layer, x, y)

    for y in range(layers_bounds[0][3]):
        for x in range(layers_bounds[0][2]):
            yield from process_tile(0, x, y)
